#pragma once

#include <dune/common/typeutilities.hh>

#include <amdis/common/Concepts.hpp>
#include <amdis/common/Logical.hpp>

namespace AMDiS
{
  namespace Traits
  {
    template <class T>
    struct IsPreGridFunction
      : std::false_type {};
  } // end namespace Traits

  namespace Concepts
  {
    /** \addtogroup Concepts
     *  @{
     **/

    namespace Definition
    {
      struct HasLocalFunction
      {
        template <class F>
        auto requires_(F&& f) -> decltype( localFunction(f) );
      };

      struct HasDerivative
      {
        template <class F>
        auto requires_(F&& f) -> decltype( derivative(f) );
      };

      struct HasLocalFunctionDerivative
      {
        template <class F>
        auto requires_(F&& f) -> decltype( derivative(localFunction(f)) );
      };

      struct HasLocalFunctionOrder
      {
        template <class F>
        auto requires_(F&& f) -> decltype( order(localFunction(f)) );
      };

      struct HasPartial
      {
        template <class F>
        auto requires_(F&& f) -> decltype( partial(f, index_<0>) );
      };

      struct HasOrder
      {
        template <class F>
        auto requires_(F&& f) -> decltype( order(f) );
      };

      struct HasGridFunctionTypes
      {
        template <class GF>
        auto requires_(GF const& /*gf*/) -> void_t<
          typename GF::Range,
          typename GF::Domain,
          typename GF::EntitySet >;
      };

      template <class F, int dow>
      constexpr bool CallableDow =
        Concepts::Callable<F, Dune::FieldVector<double, dow>>;

    } // end namespace Definition


    /// \brief GridFunction GF has free function `localFunction(GF)`
    template <class GF>
    constexpr bool HasLocalFunction = models<Definition::HasLocalFunction(GF)>;

    /// \brief GridFunction GF has free function `derivative(F)`
    template <class GF>
    constexpr bool HasDerivative = models<Definition::HasDerivative(GF)>;

    /// \brief GridFunction GF has free function `derivative(localFunction(F))`
    template <class GF>
    constexpr bool HasLocalFunctionDerivative = models<Definition::HasLocalFunctionDerivative(GF)>;

    /// \brief GridFunction GF has free function `order(localFunction(F))`
    template <class GF>
    constexpr bool HasLocalFunctionOrder = models<Definition::HasLocalFunctionOrder(GF)>;

    /// \brief Functor F has free function `partial(F,_0)`
    template <class F>
    constexpr bool HasPartial = models<Definition::HasPartial(F)>;

    /// \brief LocalFuncion LF has free function `order(F)`
    template <class LF>
    constexpr bool HasOrder = models<Definition::HasOrder(LF)>;


    /// \brief Functor F is collable with GlobalCoordinates `F(Dune::FieldVector<double,DOW>)`
#ifndef WORLDDIM
    template <class F>
    constexpr bool CallableDomain =
      Definition::CallableDow<F, 1> || Definition::CallableDow<F, 2> || Definition::CallableDow<F, 3>;
#else
    template <class F>
    constexpr bool CallableDomain =
      Definition::CallableDow<F, WORLDDIM>;
#endif

    /// \brief GridFunction GF is a Type that has LocalFunction and provides some
    /// typedefs for `Domain`, `Range`, and `EntitySet`.
    template <class GF>
    constexpr bool GridFunction =
      HasLocalFunction<GF> && models<Definition::HasGridFunctionTypes(GF)>;

    /// \brief Concept is fulfilled, if at least one of the massed Expressions
    /// can be converted to a GridFunction, or is already a GridFunction.
    template <class... GFs>
    constexpr bool AnyGridFunction =
      any_of_v<GridFunction<remove_cvref_t<GFs>>...> ||
      any_of_v<Traits::IsPreGridFunction<remove_cvref_t<GFs>>::value...>;

    /** @} **/

  } // end namespace Concepts

  template <class Traits,
    bool valid = !std::is_same<typename Traits::Range, Dune::Functions::InvalidRange>::value>
  constexpr bool isValidRange()
  {
    static_assert(valid, "Traits::Range is invalid");
    return valid;
  }


  // polynomial order of local functions
  template <class LocalFunction,
    class = void_t<decltype(std::declval<LocalFunction>().order())> >
  int order(LocalFunction const& lf)
  {
    return lf.order();
  }


#ifndef DOXYGEN
  template <class PreGridFct, class = void>
  struct GridFunctionCreator
      : PreGridFct::Creator {};

  namespace Impl
  {
    // Specialization for type that is already a GridFunction
    template <class GridFct, class GridView>
    GridFct const& makeGridFunctionImpl(GridFct const& gridFct, GridView const& /*gridView*/, std::true_type)
    {
      return gridFct;
    }

    // Use the \ref GridFunctionCreator to create a gridFunction from a preGridFunction
    template <class PreGridFct, class GridView,
              class Creator = GridFunctionCreator<PreGridFct>>
    decltype(auto) makeGridFunctionImpl(PreGridFct const& preGridFct, GridView const& gridView, std::false_type)
    {
      return Creator::create(preGridFct, gridView);
    }
  }
#endif


  /// \brief Generator for Gridfunctions from Expressions (PreGridfunctions)
  /**
   * \ingroup GridFunctions
   * Create an evaluable GridFunction from an expression that itself can not be
   * evaluated. Therefore, it binds the GridFunction to a GridView.
   *
   * **Example:**
   * ```
   * ProblemStat<Traits> prob("name");
   * prob.initialize(INIT_ALL);
   *
   * auto gridFct = makeGridFunction(Expression, prob.leafGridView());
   *
   * // eval GridFunction at GlobalCoordinates
   * auto value = gridFct(Dune::FieldVector<double,2>{1.0, 2.0});
   *
   * auto localFct = localFunction(gridFct);
   * for (auto const& element : elements(prob.leafGridView())) {
   *   localFct.bind(element);
   *   // eval LocalFunction at local coordinates
   *   auto x = localFct(element.geometry().center());
   *   localFct.unbind();
   * }
   * ```
   *
   * In contrast to Expressions, GridFunctions can be evaluated and
   * - have the free-function \ref localFunction() to obtain a LocalFunction
   * - its LocalFunctions have the free-function \ref order() to obtain the
   *   polynomial order of the Expression (if available)
   * - its LocalFunctions have the free-function \ref derivative() to
   *   differentiate the Expression with respect to global Coordinates.
   *   A derivative Expression can be created, using \ref gradientAtQP() that
   *   can be converted to a GridFunction afterwards.
   **/
  template <class PreGridFct, class GridView>
  decltype(auto) makeGridFunction(PreGridFct const& preGridFct, GridView const& gridView)
  {
    using isGridFct = bool_t<Concepts::GridFunction<PreGridFct>>;
    return Impl::makeGridFunctionImpl(preGridFct, gridView, isGridFct{});
  }

} // end namespace AMDiS
